unit FLPBD_Switch;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, ShellNotifyIcon, Vcl.Menus, Vcl.StdCtrls,
  IdBaseComponent, IdComponent, IdCustomTCPServer, IdTCPServer, IdTCPClient, IdContext, Hotkeys,
  System.Generics.Collections;

type
  TClipboardStream = class(TMemoryStream)
  private
    { Private 錾 }
    FHashKey: Integer;
    FPrevFormat: UINT;  //ClipboardFormatID
    FFileName,
    FTitle: string;
    function GetHashKey: Integer;
    function GetPrevFormat: UINT;
    function GetTitle: string;
  public
    { Public 錾 }
    constructor Create;
    procedure Clear;
    procedure LoadFromFile(const AFileName: string);
    procedure LoadFromClipboard(NewOwner: HWND);
    procedure SaveToFile(const AFileName: string);
    procedure SaveToClipboard(NewOwner: HWND);
    function HasFormats(Formats: TEnumerable<string>): Boolean;
    function FindFormat(Format: UINT): DWORD;
    function AsNames: string;
    function AsText: string;
    function AsBitmap: TBitmap;
    property HashKey: Integer read GetHashKey;
    property PrevFormat: UINT read GetPrevFormat;
    property FileName: string read FFileName write FFileName;
    property Title: string read GetTitle;
  end;

  TKeybdEvent = record
    WindowTitle: string;
    ClassName: string;
    KeyStroke:string;
  end;

  TSendKeys = class(TList<TKeybdEvent>)
  private
    { Private 錾 }
    function KeyText(const Wnd: HWND): string;
  public
    { Public 錾 }
    procedure KeybdEvent(const Receiver: HWND);
  end;

  TSendThread = class(TThread)
  private
    FStream: TClipboardStream;
    FTCPClient: TIdTCPClient;
  protected
    procedure Execute; override;
  public
    constructor Create(AData: TClipboardStream; AHost: string; APort: Word); virtual;
    destructor Destroy;  override;
  end;

const
  AM_EVENT = WM_APP + 1;

type
  TSwitchForm = class(TForm)
    CheckTimer: TTimer;
    TrayIcon: TShellNotifyIcon;
    TrayMenu: TPopupMenu;
    ConfigMnu: TMenuItem;
    TrayN1Mnu: TMenuItem;
    RestartMnu: TMenuItem;
    ClearClipbrdMnu: TMenuItem;
    ClearHistoryMnu: TMenuItem;
    TrayN2Mnu: TMenuItem;
    ExitMnu: TMenuItem;
    HistoryList: TListBox;
    PreviewImage: TImage;
    TCPServer: TIdTCPServer;
    procedure ApplicationIdle(Sender: TObject; var Done: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormDestroy(Sender: TObject);
    procedure CheckTimerTimerAtAlway(Sender: TObject);
    procedure CheckTimerTimerAtOnce(Sender: TObject);
    procedure ConfigMnuClick(Sender: TObject);
    procedure RestartMnuClick(Sender: TObject);
    procedure ClearClipbrdMnuClick(Sender: TObject);
    procedure ClearHistoryMnuClick(Sender: TObject);
    procedure ExitMnuClick(Sender: TObject);
    procedure HistoryListClick(Sender: TObject);
    procedure HistoryListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure TCPServerExecute(AContext: TIdContext);
  private
    { Private 錾 }
    FNextClipboardViewer: HWND;
    FSequenceNo: Cardinal;
    procedure UpdateClientHeight;
    procedure UpdateClientRect;
    procedure ReadOption;
    procedure WriteOption;
    function Logged: Boolean;
    procedure LoggingStart;
    procedure LoggingStop;
    procedure LoggingRestart;
    procedure UpdatePreview;
    procedure CheckClipboardChain;
    procedure SelectNextItem;
    procedure SelectPrevItem;
    procedure SelectCurrentItem;
    procedure TrimClipboardData;
    function AddClipboardData(AStream: TClipboardStream): Boolean;
    procedure SendThreadTerminate(Sender: TObject);
    procedure SendData(AStream: TClipboardStream);
    procedure PasteData;
    procedure WMHotKey(var Message: TWMHotKey); message WM_HOTKEY;
    procedure WMDisplayChange(var Message: TMessage); message WM_DISPLAYCHANGE;
    procedure WMDrawClipboard(var Message: TWMDrawClipboard); message WM_DRAWCLIPBOARD;
    procedure WMChangeCBChain(var Message: TWMChangeCBChain); message WM_CHANGECBCHAIN;
    procedure WMQueryEndSession(var Msg: TWMQueryEndSession); message WM_QUERYENDSESSION;
    procedure AMEvent(var Message: TMessage); message AM_EVENT;
  protected
    { protected 錾 }
    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public 錾 }
    HistoryMax: Integer;
    Interception: Boolean;
    ForwardKey,
    BackwardKey: TShortCut;
    HotClick: Integer;
    HistorySize,
    PreviewSize: Integer;
    PreviewFont: TFont;
    Formats: TDictionary<string, Cardinal>; //Name, ID
    SendKeys: TSendKeys;
    Connections: TThreadList<TSendThread>;
  end;

function FormatIdToName(FormatID: UINT): string;

var
  SwitchForm: TSwitchForm;

implementation

{$R *.dfm}

uses
  FLPBD_Config, FLPBD_Exit,
  System.UITypes, System.Hash, System.Math, WndInfo, System.StrUtils, AppData, SocketUtils;

type
  tagKBDLLHOOKSTRUCT = record
    vkCode: DWORD;
    scanCode: DWORD;
    flags: DWORD;
    time: DWORD;
    dwExtraInfo: ULONG_PTR;
  end;
  TKbdLLHookStruct = tagKBDLLHOOKSTRUCT;
  PKbdLLHookStruct = ^TKbdLLHookStruct;

  tagMSLLHOOKSTRUCT = record
    pt: TPoint;
    mouseData: DWORD;
    flags: DWORD;
    time: DWORD;
    dwExtraInfo: DWORD;
  end;
  TMsLLHookStruct = tagMSLLHOOKSTRUCT;
  PMsLLHookStruct = ^TMsLLHookStruct;

  TWheelGesture = (wgIdle, wgHold, wgIgnition, wgMisfire);

const
  LLKHF_EXTENDED = KF_EXTENDED shr 8;
  LLKHF_LOWER_IL_INJECTED = $00000002;
  LLKHF_INJECTED = $00000010;
  LLKHF_ALTDOWN = KF_ALTDOWN shr 8;
  LLKHF_UP = KF_UP shr 8;

  LLMHF_INJECTED = $00000001;
  LLMHF_LOWER_IL_INJECTED = $00000002;

  XBUTTON1 = $0001;
  XBUTTON2 = $0002;

  ForwardID = 1;
  BackwardID = 2;
  CurrentID = 3;

var
  KeyboardHook: HHOOK = 0;
  MouseHook: HHOOK =0;
  EventHook: THandle = 0;

  Modifiers: Word = 0;
  ForwardVK: Word = 0;
  BackwardVK: Word = 0;
  ForwardMOD: Word = 0;
  BackwardMOD: Word = 0;
  HotButton: Integer = 0;

{ Routine }

procedure DebugPrint(State, Message: string);
begin
  OutputDebugString(PChar(Format('FLPBD.%s: %s', [State, Message])));
end;

function FormatIdToName(FormatID: UINT): string;
var
  FormatName: array[0..255] of Char;
begin
  if GetClipboardFormatName(FormatID, FormatName, 255) > 0 then
    Result := FormatName
  else
    case FormatID of
      CF_TEXT:            Result := 'CF_TEXT';
      CF_BITMAP:          Result := 'CF_BITMAP';
      CF_METAFILEPICT:    Result := 'CF_METAFILEPICT';
      CF_SYLK:            Result := 'CF_SYLK';
      CF_DIF:             Result := 'CF_DIF';
      CF_TIFF:            Result := 'CF_TIFF';
      CF_OEMTEXT:         Result := 'CF_OEMTEXT';
      CF_DIB:             Result := 'CF_DIB';
      CF_PALETTE:         Result := 'CF_PALETTE';
      CF_PENDATA:         Result := 'CF_PENDATA';
      CF_RIFF:            Result := 'CF_RIFF';
      CF_WAVE:            Result := 'CF_WAVE';
      CF_UNICODETEXT:     Result := 'CF_UNICODETEXT';
      CF_ENHMETAFILE:     Result := 'CF_ENHMETAFILE';
      CF_HDROP:           Result := 'CF_HDROP';
      CF_LOCALE:          Result := 'CF_LOCALE';
      CF_MAX:             Result := 'CF_MAX';
      CF_OWNERDISPLAY:    Result := 'CF_OWNERDISPLAY';
      CF_DSPTEXT:         Result := 'CF_DSPTEXT';
      CF_DSPBITMAP:       Result := 'CF_DSPBITMAP';
      CF_DSPMETAFILEPICT: Result := 'CF_DSPMETAFILEPICT';
      CF_DSPENHMETAFILE:  Result := 'CF_DSPENHMETAFILE';
      {! vCx[gf[^GAB
      CF_PRIVATEFIRST:    Result := 'CF_PRIVATEFIRST';
      CF_PRIVATELAST:     Result := 'CF_PRIVATELAST';
      CF_GDIOBJFIRST:     Result := 'CF_GDIOBJFIRST';
      CF_GDIOBJLAST:      Result := 'CF_GDIOBJLAST';
      !}
    else
      Result := 'CF_' + IntToStr(FormatID);
    end;
end;

function GetModifyButtons: Word;
begin
  Result := 0;
  if GetKeyState(VK_LBUTTON) < 0 then Inc(Result, VK_LBUTTON);
  if GetKeyState(VK_RBUTTON) < 0 then Inc(Result, VK_RBUTTON);
  if GetKeyState(VK_MBUTTON) < 0 then Inc(Result, VK_MBUTTON);
  if GetKeyState(VK_XBUTTON1) < 0 then Inc(Result, VK_XBUTTON1);
  if GetKeyState(VK_XBUTTON2) < 0 then Inc(Result, VK_XBUTTON2);
end;

function LowLevelKeyboardProc(nCode: Integer;
  wParam: WPARAM; lParam: PKbdLLHookStruct): LRESULT; stdcall;
begin
  Result := 0;

  if nCode = HC_ACTION then
  begin
    case wParam of
      WM_KEYDOWN, WM_SYSKEYDOWN:
      begin
        case lParam.vkCode of
          VK_LSHIFT, VK_RSHIFT:
            if (Modifiers and MOD_SHIFT) = 0 then Inc(Modifiers, MOD_SHIFT);
          VK_LCONTROL, VK_RCONTROL:
            if (Modifiers and MOD_CONTROL) = 0 then Inc(Modifiers, MOD_CONTROL);
          VK_LMENU, VK_RMENU:
            if (Modifiers and MOD_ALT) = 0 then Inc(Modifiers, MOD_ALT);
          VK_LWIN, VK_RWIN:
            if (Modifiers and MOD_WIN) = 0 then Inc(Modifiers, MOD_WIN);
        end;

        if ConfigDialog.Visible and ConfigDialog.InterceptChk.Checked then
        begin
          if ConfigDialog.ForwardKeyEdt.Handle = GetFocus then
          begin
            ConfigDialog.ForwardKeyEdt.HotKey := HotKey(lParam.vkCode, Modifiers);
            ConfigDialog.ForwardKeyEdt.SelStart := Length(ConfigDialog.ForwardKeyEdt.Text);
          end
          else if ConfigDialog.BackwardKeyEdt.Handle = GetFocus then
          begin
            ConfigDialog.BackwardKeyEdt.HotKey := HotKey(lParam.vkCode, Modifiers);
            ConfigDialog.BackwardKeyEdt.SelStart := Length(ConfigDialog.BackwardKeyEdt.Text);
          end;
        end;

        if (lParam.vkCode = ForwardVK) and (Modifiers = ForwardMOD) then
        begin
          PostMessage(SwitchForm.Handle, WM_HOTKEY, ForwardID, MakeLong(Modifiers, lParam.vkCode));
          Result := 1;
        end
        else if (lParam.vkCode = BackwardVK) and (Modifiers = BackwardMOD) then
        begin
          PostMessage(SwitchForm.Handle, WM_HOTKEY, BackwardID, MakeLong(Modifiers, lParam.vkCode));
          Result := 1;
        end;
      end;
      WM_KEYUP, WM_SYSKEYUP:
      begin
        case lParam.vkCode of
          VK_LSHIFT, VK_RSHIFT:
            if (Modifiers and MOD_SHIFT) <> 0 then Dec(Modifiers, MOD_SHIFT);
          VK_LCONTROL, VK_RCONTROL:
            if (Modifiers and MOD_CONTROL) <> 0 then Dec(Modifiers, MOD_CONTROL);
          VK_LMENU, VK_RMENU:
            if (Modifiers and MOD_ALT) <> 0 then Dec(Modifiers, MOD_ALT);
          VK_LWIN, VK_RWIN:
            if (Modifiers and MOD_WIN) <> 0 then Dec(Modifiers, MOD_WIN);
        end;
      end;
    end;
  end;

  if Result = 0 then
    Result := CallNextHookEx(KeyboardHook, nCode, wParam, Winapi.Windows.LPARAM(lParam));
end;

function LowLevelMouseProc(nCode: Integer;
  wParam: WPARAM; lParam: PMsLLHookStruct): LRESULT; stdcall;
begin
  Result := 0;

  if nCode = HC_ACTION then
  begin
    case wParam of
      WM_LBUTTONDOWN, WM_LBUTTONUP:
      begin
        Modifiers := GetModifyKeys;
        PostMessage(SwitchForm.Handle, WM_HOTKEY, CurrentID, MakeLong(Modifiers, VK_LBUTTON))
      end;
      WM_RBUTTONDOWN, WM_RBUTTONUP:
      begin
        Modifiers := GetModifyKeys;
        PostMessage(SwitchForm.Handle, WM_HOTKEY, CurrentID, MakeLong(Modifiers, VK_RBUTTON))
      end;
      WM_MBUTTONDOWN, WM_MBUTTONUP:
      begin
        Modifiers := GetModifyKeys;
        PostMessage(SwitchForm.Handle, WM_HOTKEY, CurrentID, MakeLong(Modifiers, VK_MBUTTON))
      end;
      WM_XBUTTONDOWN, WM_XBUTTONUP:
      begin
        Modifiers := GetModifyKeys;
        case HiWord(lParam.mouseData) of
          XBUTTON1: PostMessage(SwitchForm.Handle, WM_HOTKEY, CurrentID, MakeLong(Modifiers, VK_XBUTTON1));
          XBUTTON2: PostMessage(SwitchForm.Handle, WM_HOTKEY, CurrentID, MakeLong(Modifiers, VK_XBUTTON2));
        end;
      end;
      WM_MOUSEWHEEL:
      begin
        if GetKeyState(HotButton) < 0 then
        begin
          Modifiers := GetModifyKeys;
          if HiWord(lParam.mouseData) = 120 then
            PostMessage(SwitchForm.Handle, WM_HOTKEY, BackwardID, MakeLong(Modifiers, HotButton))
          else
            PostMessage(SwitchForm.Handle, WM_HOTKEY, ForwardID, MakeLong(Modifiers, HotButton));
          Result := 1;
        end;
      end;
    end;
  end;

  if Result = 0 then
    Result := CallNextHookEx(0, nCode, wParam, Winapi.Windows.LPARAM(lParam));
end;

procedure WinEventProc(hWinEventHook: THandle; event: DWORD; hwnd: HWND;
  idObject, idChild: Longint; idEventThread, dwmsEventTime: DWORD); stdcall;
begin
  PostMessage(SwitchForm.Handle, AM_EVENT, idEventThread, dwmsEventTime);
end;

{ TClipboardStream }

constructor TClipboardStream.Create;
begin
  inherited Create;

  FHashKey := 0;
  FPrevFormat := 0;
  FFileName := '';
  FTitle := '';
end;

procedure TClipboardStream.Clear;
begin
  FHashKey := 0;
  FPrevFormat := 0;
end;

procedure TClipboardStream.LoadFromFile(const AFileName: string);
begin
  inherited LoadFromFile(AFileName);

  FFileName := AFileName;
end;

procedure TClipboardStream.SaveToFile(const AFileName: string);
begin
  inherited SaveToFile(AFileName);

  FFileName := AFileName;
end;

function TClipboardStream.GetHashKey: Integer;
begin
  if FHashKey = 0 then
    FHashKey := THashBobJenkins.GetHashValue(Memory^, Size, 0);
  Result := FHashKey;
end;

function TClipboardStream.GetPrevFormat: UINT;
var
  FormatID: UINT;
  DataSize: DWORD;
begin
  if FPrevFormat = 0 then
  begin
    Position := 0;
    while Read(FormatID, SizeOf(FormatID)) = SizeOf(FormatID) do
    begin
      if FormatID in [CF_TEXT, CF_DIB] then
      begin
        FPrevFormat := FormatID;
        Break;
      end
      else if Position = 0 then
        FPrevFormat := FormatID;

      Read(DataSize, SizeOf(DataSize));
      Seek(DataSize, soCurrent);
    end;
  end;
  Result := FPrevFormat;
end;

function TClipboardStream.GetTitle: string;
var
  s: string;
  i: Integer;
begin
  if FTitle = '' then
  begin
    case PrevFormat of
      CF_TEXT: s := TrimLeft(AsText);
      CF_DIB: s := '(BITMAP)';
    else
      s := AsNames;
    end;

    i := Min(Pos(#13, s), Pos(#10, s));
    if i > 0 then
      s := Copy(s, 0, i);

    if s = '' then
      FTitle := '(BLANK TEXT)'
    else
      FTitle := s;
  end;
  Result := FTitle;
end;

procedure TClipboardStream.LoadFromClipboard(NewOwner: HWND);
var
  FormatID: UINT;
  DataHnd: THandle;
  DataSize: DWORD;
  Buffer: Pointer;
begin
  Clear;

  if OpenClipboard(NewOwner) then
  begin
    try
      FormatID := EnumClipboardFormats(0);
      while FormatID > 0 do
      begin
        Write(FormatID, SizeOf(FormatID));

        DataHnd := GetClipboardData(FormatID);
        if DataHnd <> 0 then
        begin
          Buffer := GlobalLock(DataHnd);
          try
            DataSize := GlobalSize(DataHnd);
            Write(DataSize, SizeOf(DataSize));
            Write(Buffer^, DataSize);
          finally
            GlobalUnlock(DataHnd);
          end;
        end
        else
        begin
          DataSize := 0;
          Write(DataSize, SizeOf(DataSize));
        end;
        FormatID := EnumClipboardFormats(FormatID);
      end;
      {$IFDEF DEBUG}
      DebugPrint('TClipboardStream', 'ReadClipboard');
      {$ENDIF DEBUG}
    finally
      CloseClipboard;
    end;
  end
  else
    DebugPrint('TClipboardStream', 'UnreadClipboard');
end;

procedure TClipboardStream.SaveToClipboard(NewOwner: HWND);
var
  FormatID: UINT;
  DataSize: DWORD;
  DataHnd: THandle;
  Buffer: Pointer;
begin
  if OpenClipboard(NewOwner) then
  begin
    try
      EmptyClipboard;

      Position := 0;
      while Read(FormatID, SizeOf(FormatID)) = SizeOf(FormatID) do
      begin
        Read(DataSize, SizeOf(DataSize));
        DataHnd := GlobalAlloc(GHND, DataSize);

        Buffer := GlobalLock(DataHnd);
        Read(Buffer^, DataSize);
        GlobalUnlock(DataHnd);

        SetClipboardData(FormatID, DataHnd);
        //! GlobalFree(DataHnd);  J̓Nbv{[hsB
      end;
      {$IFDEF DEBUG}
      DebugPrint('TClipboardStream', 'WriteClipboard');
      {$ENDIF DEBUG}
    finally
      CloseClipboard;
    end;
  end
  else
    DebugPrint('TClipboardStream', 'UnwriteClipboard');
end;

function TClipboardStream.HasFormats(Formats: TEnumerable<string>): Boolean;
var
  FormatID: UINT;
  FormatName: string;
  DataSize: DWORD;
begin
  Result := False;

  Position := 0;
  while Read(FormatID, SizeOf(FormatID)) = SizeOf(FormatID) do
  begin
    for FormatName in Formats do
    begin
      Result := FormatName = FormatIdToName(FormatID);

      if Result then Exit;
    end;
    Read(DataSize, SizeOf(DataSize));
    Seek(DataSize, soCurrent);
  end;
end;

function TClipboardStream.FindFormat(Format: UINT): DWORD;
var
  FormatID: UINT;
  DataSize: DWORD;
begin
  Result := 0;

  Position := 0;
  while Read(FormatID, SizeOf(FormatID)) = SizeOf(FormatID) do
  begin
    Read(DataSize, SizeOf(DataSize));

    if FormatID = Format then
    begin
      Result := DataSize;
      Exit;
    end;
    Seek(DataSize, soCurrent);
  end;
end;

function TClipboardStream.AsNames: string;
var
  FormatID: UINT;
  DataSize: DWORD;
begin
  Result := '';

  Position := 0;
  while Read(FormatID, SizeOf(FormatID)) = SizeOf(FormatID) do
  begin
    Result := Format('%s(%s)' + #13#10, [Result, FormatIdToName(FormatID)]);

    Read(DataSize, SizeOf(DataSize));
    Seek(DataSize, soCurrent);
  end;
  if Result = '' then
    Result := '(NULL)';
end;

function TClipboardStream.AsText: string;
var
  DataSize: DWORD;
  Buffer: RawByteString;
begin
  Result := '';

  DataSize := FindFormat(CF_TEXT);
  if DataSize > 0 then
  begin
    SetLength(Buffer, DataSize);
    Read(Pointer(Buffer)^, DataSize);
    Result := string(Buffer);
  end;
end;

function TClipboardStream.AsBitmap: TBitmap;
var
  DataSize: DWORD;
  Buffer: TMemoryStream;
  BmpHeader: TBitmapFileHeader;
begin
  Result := TBitmap.Create;

  DataSize := FindFormat(CF_DIB);
  if DataSize > 0 then
  begin
    Buffer := TMemoryStream.Create;
    try
      BmpHeader.bfType := $4D42;  //͐ݒ肵ȂĂC݂B
      Buffer.Write(BmpHeader, SizeOf(TBitmapFileHeader));
      Buffer.CopyFrom(Self, DataSize);
      Buffer.Position := 0;
      Result.LoadFromStream(Buffer);
    finally
      Buffer.Free;
    end;
  end;
end;

{ TSendKeys }

function TSendKeys.KeyText(const Wnd: HWND): string;
var
  i: Integer;
begin
  Result := 'Ctrl+V';

  for i := 0 to Count - 1 do
    if MatchesWindow(Wnd, Items[i].ClassName, Items[i].WindowTitle) then
    begin
      Result := Items[i].KeyStroke;
      Break;
    end;
end;

procedure TSendKeys.KeybdEvent(const Receiver: HWND);
var
  Keys: TList<Word>;
  Stroke,
  KeybdEventStr,
  ShortCutStr,
  s: string;
  Key: TShortCut;
  i: Integer;
begin
  if Receiver <> GetForegroundWindow then
    SetForegroundWindow(Receiver);

  Keys := TList<Word>.Create;
  try
    Stroke := KeyText(Receiver);
    for KeybdEventStr in SplitString(Stroke, '|') do
    begin
      Keys.Clear;

      s := KeybdEventStr;
      while True do
      begin
        if CompareFront(s, HotKeyCaps[hkcShift]) then Keys.Add(VK_SHIFT)
        else if CompareFront(s, HotKeyCaps[hkcCtrl]) then Keys.Add(VK_CONTROL)
        else if CompareFront(s, HotKeyCaps[hkcAlt]) then Keys.Add(VK_MENU)
        else if CompareFront(s, HotKeyCaps[hkcWin]) then Keys.Add(VK_LWIN)
        else Break;
      end;
      if s <> '' then
      begin
        for ShortCutStr in SplitString(s, '+') do
          for Key := $08 to $255 do
            if AnsiSameText(ShortCutStr, HotKeyToText(Key)) then
            begin
              Keys.Add(Key);
              Break;
            end;
      end;
      for i := 0 to Keys.Count - 1 do
      begin
        Sleep(10);
        keybd_event(Keys.Items[i], 0, 0, 0);
      end;
      for i := Keys.Count - 1 downto 0 do
      begin
        Sleep(10);
        keybd_event(Keys.Items[i], 0, KEYEVENTF_KEYUP, 0);
      end;
    end;
  finally
    Keys.Free;
  end;
end;

{ TSendThread }

constructor TSendThread.Create(AData: TClipboardStream; AHost: string; APort: Word);
begin
  FStream := TClipboardStream.Create;
  FStream.LoadFromStream(AData);
  FTCPClient := TIdTCPClient.Create(nil);
  FTCPClient.Host := AHost;
  FTCPClient.Port := APort;

  inherited Create(True);
end;

procedure TSendThread.Execute;
begin
  try
    FTCPClient.Connect;
    try
      try
        FTCPClient.IOHandler.Write(FStream.Size);
        FTCPClient.IOHandler.WriteBufferOpen;
        FTCPClient.IOHandler.Write(FStream, 0, False);
        FTCPClient.IOHandler.WriteBufferFlush;
      finally
        FTCPClient.IOHandler.WriteBufferClose;
      end;
      {$IFDEF DEBUG}
      DebugPrint('TSendThread', Format('Send(%s)%dbyte', [FTCPClient.Host, FStream.Size]));
      {$ENDIF DEBUG}
    finally
      FTCPClient.Disconnect;
    end;
  except
    DebugPrint('TSendThread', Format('Send(%s)error', [FTcpClient.Host]));
  end;
end;

destructor TSendThread.Destroy;
begin
  FStream.Free;
  FTcpClient.Free;

  inherited Destroy;
end;

{ TSwitchForm }

procedure TSwitchForm.UpdateClientHeight;
begin
  ClientHeight := Min(
    Max(PreviewImage.Width, (HistoryList.ItemHeight * HistoryList.Items.Count) + 4),
    Monitor.WorkareaRect.Height - 16
  );
end;

procedure TSwitchForm.UpdateClientRect;
begin
  if PreviewImage.Visible then
    ClientWidth := PreviewSize + HistorySize
  else
    ClientWidth := HistorySize;
  PreviewImage.Width := PreviewSize;

  UpdateClientHeight;
end;

procedure TSwitchForm.ReadOption;
var
  IniFile: TAppIniFile;
  Idents,
  Values: TStrings;
  i: Integer;
  KeybdEvent: TKeybdEvent;
  FileName: string;
  Stream: TClipboardStream;
begin
  IniFile := TAppIniFile.Create(AppDataName('ShaggyDog\Flipboard\Flipboard.ini'));
  try
    //Work
    HistoryMax := IniFile.ReadInteger('Work', 'HistoryMax', 16);

    Interception := IniFile.ReadBool('Work', 'Interception', False);
    ForwardKey := IniFile.ReadInteger('Work', 'ForwardKey', 33008);   //Alt+CapsLock
    BackwardKey := IniFile.ReadInteger('Work', 'BackwardKey', 41200); //Shift+Alt+CapsLock
    HotClick := IniFile.ReadInteger('Work', 'HotClick', VK_LBUTTON);

    TCPServer.DefaultPort := IniFile.ReadInteger('Work', 'Port', 9801);
    TCPServer.Active := IniFile.ReadBool('Work', 'Share', False);

    //View
    HistorySize := IniFile.ReadInteger('History', 'Size', 320);
    HistoryList.Font.Name := IniFile.ReadString('History', 'FontName', 'lr SVbN');
    HistoryList.Font.Size := IniFile.ReadInteger('History', 'FoneSize', 12);
    HistoryList.Font.Style := TFontStyles(Byte(IniFile.ReadInteger('History', 'FontStyle', 0)));

    PreviewImage.Visible := IniFile.ReadBool('Preview', 'Visible', True);
    PreviewSize := IniFile.ReadInteger('Preview', 'Size', 160);
    PreviewFont.Name := IniFile.ReadString('Preview', 'FontName', 'Tahoma');
    PreviewFont.Size := IniFile.ReadInteger('Preview', 'FontSize', 8);
    PreviewFont.Style := TFontStyles(Byte(IniFile.ReadInteger('Preview', 'FontStyle', 0)));

    //Formats
    Values := TStringList.Create;
    try
      IniFile.ReadSectionValues('Formats', Values);
      for i := 0 to Values.Count - 1 do
        Formats.Add(Values.Names[i], StrToInt(Values.ValueFromIndex[i]));
      if Formats.Count = 0 then
      begin
        Formats.Add('CF_TEXT', 1);
        Formats.Add('CF_BITMAP', 2);
        Formats.Add('CF_HDROP', 15);
      end;
    finally
      Values.Free;
    end;

    //SendKeys
    Idents := TStringList.Create;
    try
      IniFile.ReadSection('SendKeys', Idents);
      Values := TStringList.Create;
      try
        Values.StrictDelimiter := True;
        for i := 0 to Idents.Count - 1 do
        begin
          Values.CommaText := IniFile.ReadString('SendKeys', Idents.Strings[i], '');
          KeybdEvent.WindowTitle := Values.Strings[0];
          KeybdEvent.ClassName := Values.Strings[1];
          KeybdEvent.KeyStroke := Values.Strings[2];
          SendKeys.Add(KeybdEvent);
        end;
      finally
        Values.Free;
      end;
    finally
      Idents.Free;
    end;

    //Lock
    Idents := TStringList.Create;
    try
      IniFile.ReadSection('Lock', Idents);
      for i := 0 to Idents.Count - 1 do
      begin
        FileName := IniFile.ReadString('Lock', Idents.Strings[i], '');
        if FileExists(FileName) then
        begin
          Stream := TClipboardStream.Create;
          Stream.LoadFromFile(FileName);
          HistoryList.Items.AddObject(Stream.Title, Stream);
        end;
      end;
    finally
      Idents.Free;
    end;

    UpdateClientRect;
  finally
    IniFile.Free;
  end;
end;

procedure TSwitchForm.WriteOption;
var
  IniFile: TAppIniFile;
  i: Integer;
  s: string;
  Stream: TClipboardStream;
begin
  IniFile := TAppIniFile.Create(AppDataName('ShaggyDog\Flipboard\Flipboard.ini'));
  try
    //Work
    IniFile.WriteInteger('Work', 'HistoryMax', HistoryMax);

    IniFile.WriteBool('Work', 'Interception', Interception);
    IniFile.WriteInteger('Work', 'ForwardKey', ForwardKey);
    IniFile.WriteInteger('Work', 'BackwardKey', BackwardKey);
    IniFile.WriteInteger('Work', 'HotClick', HotClick);

    IniFile.WriteBool('Work', 'Share', TCPServer.Active);
    IniFile.WriteInteger('Work', 'Port', TCPServer.DefaultPort);

    //View
    IniFile.WriteInteger('History', 'Size', HistorySize);
    IniFile.WriteString('History', 'FontName', HistoryList.Font.Name);
    IniFile.WriteInteger('History', 'FoneSize', HistoryList.Font.Size);
    IniFile.WriteInteger('History', 'FontStyle', Byte(HistoryList.Font.Style));

    IniFile.WriteBool('Preview', 'Visible', PreviewImage.Visible);
    IniFile.WriteInteger('Preview', 'Size', PreviewSize);
    IniFile.WriteString('Preview', 'FontName', PreviewFont.Name);
    IniFile.WriteInteger('Preview', 'FoneSize', PreviewFont.Size);
    IniFile.WriteInteger('Preview', 'FontStyle', Byte(PreviewFont.Style));

    //Formats
    IniFile.DeleteSection('Formats');
    for s in Formats.Keys do
      IniFile.WriteInteger('Formats', s, Formats.Items[s]);

    //SendKeys
    IniFile.DeleteSection('SendKeys');
    for i := 0 to SendKeys.Count - 1 do
    begin
      s := Format('%s,%s,%s', [
        AnsiQuotedStr(SendKeys.Items[i].WindowTitle, '"'),
        AnsiQuotedStr(SendKeys.Items[i].ClassName, '"'),
        AnsiQuotedStr(SendKeys.Items[i].KeyStroke, '"')
      ]);
      IniFile.WriteQuotedString('SendKeys', IntToStr(i), s);
    end;

    //Lock
    IniFile.DeleteSection('Lock');
    for i := 0 to HistoryList.Items.Count - 1 do
    begin
      Stream := TClipboardStream(HistoryList.Items.Objects[i]);
      if Stream.FileName <> '' then
      begin
        Stream.SaveToFile(Stream.FileName);
        IniFile.WriteQuotedString('Lock', IntToStr(i), Stream.FileName);
      end;
    end;
  finally
    IniFile.Free;
  end;
end;

function TSwitchForm.Logged: Boolean;
begin
  Result := Assigned(CheckTimer.OnTimer);
end;

procedure TSwitchForm.LoggingStart;
begin
  if Logged then Exit;

  //RegisterPopupKey
  if ForwardKey <> 0 then
    HotKeyToKey(ForwardKey, ForwardVK, ForwardMOD);
  if BackwardKey <> 0 then
    HotKeyToKey(BackwardKey, BackwardVK, BackwardMOD);
  if Interception then
  begin
    KeyboardHook := SetWindowsHookEx(WH_KEYBOARD_LL, @LowLevelKeyboardProc, hInstance, 0);
    if KeyboardHook = 0 then
      MessageDlg('WH_KEYBOARD_LL' + #13#10 + SysErrorMessage(GetLastError), mtError, [mbOk], 0);
  end
  else
  begin
    if ForwardKey <> 0 then
    begin
      if not RegisterHotKey(Handle, ForwardID, ForwardMOD, ForwardVK) then
        MessageDlg(HotKeyToText(ForwardKey) + #13#10 + SysErrorMessage(GetLastError), mtError, [mbOk], 0);
    end;
    if BackwardKey <> 0 then
    begin
      if not RegisterHotKey(Handle, BackwardID, BackwardMOD, BackwardVK) then
        MessageDlg(HotKeyToText(BackwardKey) + #13#10 + SysErrorMessage(GetLastError), mtError, [mbOk], 0);
    end;
  end;

  //InstallMouseHook
  HotButton := HotClick;
  if HotButton <> 0 then
  begin
    MouseHook := SetWindowsHookEx(WH_MOUSE_LL, @LowLevelMouseProc, hInstance, 0);
    if MouseHook = 0 then
      MessageDlg('WH_MOUSE_LL' + #13#10 + SysErrorMessage(GetLastError), mtError, [mbOk], 0);
  end;

  //InstallEventHook
  EventHook := SetWinEventHook(
    EVENT_OBJECT_DESTROY, EVENT_OBJECT_DESTROY,
    0, WinEventProc, 0, 0,
    WINEVENT_OUTOFCONTEXT or WINEVENT_SKIPOWNPROCESS or WINEVENT_SKIPOWNTHREAD
  );
  if EventHook = 0 then
  begin
    CheckTimer.OnTimer := CheckTimerTimerAtAlway;
    CheckTimer.Interval := 60000;
    CheckTimer.Enabled := True;
  end
  else
  begin
    CheckTimer.OnTimer := CheckTimerTimerAtOnce;
    CheckTimer.Interval := 3000;
    CheckTimer.Enabled := False;
  end;

  //RegisterClipboardChain
  FNextClipboardViewer := SetClipboardViewer(Handle);
end;

procedure TSwitchForm.LoggingStop;
begin
  //UnregisterClipboardChain
  ChangeClipboardChain(Handle, FNextClipboardViewer);

  //UninstallEventHook
  CheckTimer.OnTimer := nil;
  if EventHook <> 0 then
  begin
    UnhookWinEvent(EventHook);
    EventHook := 0;
  end;

  //UninstallMouseHook;
  if MouseHook <> 0 then
  begin
    UnhookWindowsHookEx(MouseHook);
    MouseHook := 0;
    HotButton := 0;
  end;

  //UnregisterPoupKey
  if KeyboardHook <> 0 then
  begin
    UnhookWindowsHookEx(KeyboardHook);
    KeyboardHook := 0;
  end;
  ForwardVK := 0;
  ForwardMOD := 0;
  BackwardVK := 0;
  BackwardMOD := 0;
end;

procedure TSwitchForm.LoggingRestart;
begin
  LoggingStop;
  try
    Application.ProcessMessages;

    Position := poScreenCenter;
  finally
    LoggingStart;
  end;
end;

procedure TSwitchForm.ApplicationIdle(Sender: TObject; var Done: Boolean);
begin
  //HistoryList.OnKeyUp ȂƂ̂ŁAAChCxgł`FbNB
  SelectCurrentItem;
end;

procedure TSwitchForm.CreateParams(var Params: TCreateParams);
begin
  inherited;

  //{[_[gɂB
  Params.Style := Params.Style and not (WS_DLGFRAME or WS_THICKFRAME) or WS_POPUP;
  //etBet Delphi10.3 ̃e[}LȏꍇAy[XgFB
  //! Params.WindowClass.Style := Params.WindowClass.Style or CS_DROPSHADOW;
  //^XNo[{^BApplication.MainFormOnTaskbar = True ł邱ƁB
  Params.ExStyle := Params.ExStyle and not WS_EX_APPWINDOW;
  Params.WndParent := Application.Handle;
end;

procedure TSwitchForm.FormCreate(Sender: TObject);
var
  Info: TWndClassEx;
begin
  PreviewFont := TFont.Create;
  Formats := TDictionary<string, Cardinal>.Create;
  SendKeys := TSendKeys.Create;
  Connections := TThreadList<TSendThread>.Create;

  ReadOption;

  Application.ShowMainForm := False;
  Application.OnDeactivate := FormDeactivate;
  //ACRgCACRɊ蓖ĂB
  GetClassInfoEx(hInstance, PChar(Application.ClassName), Info);
  TrayIcon.Icon.Handle := Info.hIconSm;
  TrayIcon.Show;

  LoggingStart;
end;

procedure TSwitchForm.UpdatePreview;
const
  TextFormat = DT_EDITCONTROL or DT_WORDBREAK;
var
  Prev, Buf: TBitmap;
  ClipRect: TRect;
  Stream: TClipboardStream;
  Txt: string;
  Scale: Double;
begin
  if not PreviewImage.Visible then Exit;

  Prev := TBitmap.Create;
  try
    Prev.Width := PreviewSize;
    Prev.Height := PreviewSize;
    Prev.Canvas.Font := PreviewFont;
    ClipRect := Prev.Canvas.ClipRect;

    Stream := TClipboardStream(HistoryList.Items.Objects[HistoryList.ItemIndex]);
    case Stream.PrevFormat of
      CF_TEXT:
      begin
        Txt := Stream.AsText;
        DrawText(Prev.Canvas.Handle, PChar(Txt), -1, ClipRect, TextFormat);
      end;
      CF_DIB:
      begin
        Buf := Stream.AsBitmap;
        try
          if (Buf.Width > PreviewSize) or (Buf.Height > PreviewSize) then
          begin
            Scale := Min(PreviewSize / Buf.Width, PreviewSize / Buf.Height);
            Prev.Width := Trunc(Buf.Width * Scale);
            Prev.Height := Trunc(Buf.Height * Scale);

            SetStretchBltMode(Prev.Canvas.Handle, HALFTONE);
            StretchBlt(
              Prev.Canvas.Handle, 0, 0, Prev.Width, Prev.Height,
              Buf.Canvas.Handle, 0, 0, Buf.Width, Buf.Height,
              SRCCOPY
            );
          end
          else
            Prev.Assign(Buf);
        finally
          Buf.Free;
        end;
      end;
    else
      Txt := Stream.AsNames;
      DrawText(Prev.Canvas.Handle, PChar(Txt), -1, ClipRect, TextFormat);
    end;
    PreviewImage.Picture.Bitmap.Assign(Prev);
  finally
    Prev.Free;
  end;
end;

procedure TSwitchForm.CheckClipboardChain;
begin
  if FSequenceNo <> GetClipboardSequenceNumber then
  begin
    ChangeClipboardChain(Handle, FNextClipboardViewer);
    FNextClipboardViewer := SetClipboardViewer(Handle);

    {$IFDEF DEBUG}
    DebugPrint('TSwitchForm', 'UpdateClipboardChain');
    {$ENDIF DEBUG}
  end;
end;

procedure TSwitchForm.SelectNextItem;

  procedure IncItemIndex(Index: Integer);
  begin
    if HistoryList.Items.Count > 0 then
    begin
      Inc(Index);
      if Index >= HistoryList.Items.Count then
        Index := 0;
      HistoryList.Selected[Index] := True;
      UpdatePreview;
    end;
  end;

begin
  if Visible then
  begin
    SetForegroundWindow(Handle);
    IncItemIndex(HistoryList.ItemIndex);
  end
  else
  begin
    CheckClipboardChain;
    IncItemIndex(-1);
    Show;
    SetForegroundWindow(Handle);
  end;
end;

procedure TSwitchForm.SelectPrevItem;

  procedure DecItemIndex(Index: Integer);
  begin
    if HistoryList.Items.Count > 0 then
    begin
      Dec(Index);
      if Index < 0 then
        Index := HistoryList.Items.Count - 1;
      HistoryList.Selected[Index] := True;
      UpdatePreview;
    end;
  end;

begin
  if Visible then
  begin
    SetForegroundWindow(Handle);
    DecItemIndex(HistoryList.ItemIndex);
  end
  else
  begin
    CheckClipboardChain;
    DecItemIndex(HistoryList.Items.Count);
    Show;
    SetForegroundWindow(Handle);
  end;
end;

procedure TSwitchForm.SelectCurrentItem;
var
  ModKey,
  ModBtn: Word;
begin
  if Visible then
  begin
    ModKey := GetModifyKeys;
    ModBtn := GetModifyButtons;
    if (ForwardMOD <> ModKey) and (BackwardMOD <> ModKey)
      and ((HotClick = 0) or (HotClick <> ModBtn)) then
    begin
      try
        if (0 = ModKey) and (0 = ModBtn) then
          PasteData;
      finally
        Hide;
      end;
    end;
  end;
end;

procedure TSwitchForm.TrimClipboardData;
var
  i: Integer;
begin
  for i := HistoryList.Items.Count - 1 downto 0 do
  begin
    if HistoryList.Items.Count <= HistoryMax then Break;

    if TClipboardStream(HistoryList.Items.Objects[i]).FileName = '' then
    begin
      TClipboardStream(HistoryList.Items.Objects[i]).Free;
      HistoryList.Items.Delete(i);
      {$IFDEF DEBUG}
      DebugPrint('TSwitchForm', 'TrimData');
      {$ENDIF DEBUG}
    end;
  end;
end;

function TSwitchForm.AddClipboardData(AStream: TClipboardStream): Boolean;
var
  i, Index: Integer;
begin
  Result := False;

  TrayIcon.Hint := AStream.Title;
  if AStream.Size = 0 then
  begin
    {$IFDEF DEBUG}
    DebugPrint('TSwitchForm', 'NullData');
    {$ENDIF DEBUG}
  end
  else if not AStream.HasFormats(Formats.Keys) then
  begin
    {$IFDEF DEBUG}
    DebugPrint('TSwitchForm', 'ExceptData');
    {$ENDIF DEBUG}
  end
  else
  begin
    Index := -1;
    for i := 0 to HistoryList.Count - 1 do
      if AStream.HashKey = TClipboardStream(HistoryList.Items.Objects[i]).HashKey then
      begin
        Index := i;
        Break;
      end;

    if Index >= 0 then
    begin
      HistoryList.Items.Move(Index, 0);
      {$IFDEF DEBUG}
      DebugPrint('TSwitchForm', 'RegisteredData');
      {$ENDIF DEBUG}
    end
    else
    begin
      HistoryList.Items.InsertObject(0, AStream.Title, AStream);
      TrimClipboardData;
      UpdateClientHeight;
      {$IFDEF DEBUG}
      DebugPrint('TSwitchForm', 'RegisterData');
      {$ENDIF DEBUG}

      Result := True;
    end;
  end;
end;

procedure TSwitchForm.SendThreadTerminate(Sender: TObject);
begin
  Connections.Remove(TSendThread(Sender));
end;

procedure TSwitchForm.SendData(AStream: TClipboardStream);
var
  HostList: TStrings;
  Host: string;
  Thread: TSendThread;
begin
  HostList := TStringList.Create;
  try
    EnumServer(SV_TYPE_NT, HostList);
    for Host in HostList do
    begin
      if Host = LocalHostName then Continue;

      Thread := TSendThread.Create(AStream, Host, TCPServer.DefaultPort);
      Connections.Add(Thread);
      Thread.FreeOnTerminate := True;
      Thread.OnTerminate := SendThreadTerminate;
      Thread.Start;
    end;
  finally
    HostList.Free;
  end;
end;

procedure TSwitchForm.PasteData;
var
  Index: Integer;
  Wnd: HWND;
begin
  Hide;

  Index := HistoryList.ItemIndex;
  if Index >= 0 then
  begin
    TClipboardStream(HistoryList.Items.Objects[Index]).SaveToClipboard(Handle);
    HistoryList.Items.Move(Index, 0);
    TrayIcon.Hint := HistoryList.Items.Strings[0];

    Wnd := GetForegroundWindow;
    if Wnd > 0 then
    begin
      {$IFDEF DEBUG}
      DebugPrint('PasteTo', Format('%s,"%s"',[GetClassNameStr(Wnd), GetWindowTextStr(Wnd)]));
      {$ENDIF DEBUG}
      Application.Minimize;
      try
        SendKeys.KeybdEvent(Wnd);
      finally
        Application.Restore;
      end;
    end;
  end;
end;

procedure TSwitchForm.WMHotKey(var Message: TWMHotKey);
begin
  if Message.HotKey = ForwardID then
    SelectNextItem
  else if Message.HotKey = BackwardID then
    SelectPrevItem
  else if Message.HotKey = CurrentID then
    SelectCurrentItem;
end;

procedure TSwitchForm.WMDisplayChange(var Message: TMessage);
begin
  Monitor;  //TScreen.Monitors ̕s΍B͎Ă݂ǁAꉞB

  LoggingRestart;
end;

procedure TSwitchForm.WMDrawClipboard(var Message: TWMDrawClipboard);
var
  Added: Boolean;
  Stream: TClipboardStream;
begin
  try
    FSequenceNo := GetClipboardSequenceNumber;
    if Handle = GetClipboardOwner then Exit;

    if Logged then
    begin
      {$IFDEF DEBUG}
      DebugPrint('TSwitchForm', 'WMDrawClipboard');
      {$ENDIF DEBUG}
      Added := False;
      Stream := TClipboardStream.Create;
      try
        Stream.LoadFromClipboard(Handle);
        Added := AddClipboardData(Stream);
        if TCPServer.Active and (Stream.Size > 0) then
          SendData(Stream);
      finally
        if not Added then Stream.Free;
      end;
    end;
  finally
    if FNextClipboardViewer > 0 then
      SendMessage(FNextClipboardViewer, WM_DRAWCLIPBOARD, 0, 0);
  end;
end;

procedure TSwitchForm.WMChangeCBChain(var Message: TWMChangeCBChain);
begin
  if Message.Remove = FNextClipboardViewer then
  begin
    FNextClipboardViewer := Message.Next;
    Message.Result := 0;
  end
  else
    SendMessage(FNextClipboardViewer, WM_CHANGECBCHAIN, Message.Remove, Message.Next);

  {$IFDEF DEBUG}
  DebugPrint('TSwitchForm', 'WMChangeCBChain');
  {$ENDIF DEBUG}
end;

procedure TSwitchForm.WMQueryEndSession(var Msg: TWMQueryEndSession);
begin
  Close;
end;

procedure TSwitchForm.AMEvent(var Message: TMessage);
begin
  inherited;

  //^C}[ĐݒB
  CheckTimer.Enabled := False;
  CheckTimer.Enabled := True;
end;

procedure TSwitchForm.FormShow(Sender: TObject);
begin
  Application.OnIdle := ApplicationIdle;
end;

procedure TSwitchForm.FormHide(Sender: TObject);
begin
  Application.OnIdle := nil;
end;

procedure TSwitchForm.FormDeactivate(Sender: TObject);
begin
  if SwitchForm.Visible then
    SwitchForm.Hide;
end;

procedure TSwitchForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  LockList: TList<TSendThread>;
  Connected: Boolean;
  i: Integer;
begin
  LockList := Connections.LockList;
  try
    Connected := LockList.Count > 0;
  finally
    Connections.UnlockList;
  end;
  if Connected then
  begin
    MessageBeep(MB_ICONEXCLAMATION);
    CanClose :=  ExitDialog.ShowModal = mrOk;
  end;

  if CanClose then
  begin
    LoggingStop;

    TCPServer.Active := False;
    CheckTimer.Enabled := False;
    LockList := Connections.LockList;
    try
      for i := 0 to LockList.Count - 1 do
      begin
        LockList.Items[i].OnTerminate := nil;
        LockList.Items[i].Terminate;
      end;
    finally
      Connections.UnlockList;
    end;
  end;
end;

procedure TSwitchForm.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to HistoryList.Items.Count - 1 do
  begin
    if Assigned(HistoryList.Items.Objects[i]) then
      TClipboardStream(HistoryList.Items.Objects[i]).Free;
  end;
  PreviewFont.Free;
  Formats.Free;
  SendKeys.Free;
  Connections.Free;
end;

procedure TSwitchForm.CheckTimerTimerAtAlway(Sender: TObject);
begin
  CheckClipboardChain;
end;

procedure TSwitchForm.CheckTimerTimerAtOnce(Sender: TObject);
begin
  CheckTimer.Enabled := False;
  CheckClipboardChain;
end;

procedure TSwitchForm.ConfigMnuClick(Sender: TObject);
var
  Unhooked: Boolean;
begin
  TrayIcon.Hide;
  try
    Unhooked := KeyboardHook = 0;
    if Unhooked then
      KeyboardHook := SetWindowsHookEx(WH_KEYBOARD_LL, @LowLevelKeyboardProc, hInstance, 0);
    try
      if ConfigDialog.ShowModal = mrOk then
      begin
        WriteOption;
        TrimClipboardData;
        UpdateClientRect;
      end;
    finally
      if Unhooked then
      begin
        UnhookWindowsHookEx(KeyboardHook);
        KeyboardHook := 0;
      end;
    end;
    LoggingRestart;
  finally
    TrayIcon.Show;
  end;
end;

procedure TSwitchForm.RestartMnuClick(Sender: TObject);
begin
  LoggingRestart;
end;

procedure TSwitchForm.ClearClipbrdMnuClick(Sender: TObject);
begin
  if OpenClipboard(Handle) then
  begin
    try
      EmptyClipboard;
      TrayIcon.Hint := '(NULL)';
    finally
      CloseClipboard;
    end;
  end
  else
    raise Exception.Create(SysErrorMessage(GetLastError));
end;

procedure TSwitchForm.ClearHistoryMnuClick(Sender: TObject);
var
  i: Integer;
  Stream: TClipboardStream;
begin
  for i := HistoryList.Items.Count - 1 downto 0 do
  begin
    Stream := TClipboardStream(HistoryList.Items.Objects[i]);
    if Stream.FileName = '' then
    begin
      Stream.Free;
      HistoryList.Items.Delete(i);
    end;
  end;
  UpdateClientHeight;
end;

procedure TSwitchForm.ExitMnuClick(Sender: TObject);
begin
  Close;
end;

procedure TSwitchForm.HistoryListClick(Sender: TObject);
begin
  UpdatePreview;
end;

procedure TSwitchForm.HistoryListKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  //Ô߁AG^[ƃGXP[vB
  case Key of
    VK_RETURN: SelectCurrentItem;
    VK_ESCAPE: Close;
  else
    SelectCurrentItem;
  end;
end;

procedure TSwitchForm.TCPServerExecute(AContext: TIdContext);
var
  Added: Boolean;
  Stream: TClipboardStream;
  ServerName: string;
  RecSize: LongInt;
begin
  Added := False;
  Stream := TClipboardStream.Create;
  try
    ServerName := GetServerName(AContext.Binding.PeerIP);
    AContext.Connection.IOHandler.ReadTimeout := 3000;
    try
      RecSize := AContext.Connection.IOHandler.ReadInt32;
      AContext.Connection.IOHandler.ReadStream(Stream, RecSize);
    except
      DebugPrint('TSwitchForm', Format('Receive(%s)error', [ServerName]));
      Exit;
    end;
    {$IFDEF DEBUG}
    DebugPrint('TSwitchForm', Format('Receive(%s)%dbyte', [ServerName, RecSize]));
    {$ENDIF DEBUG}

    Added := AddClipboardData(Stream);
  finally
    if not Added then Stream.Free;
  end;
end;

end.
