(*  WndInfo

Ver, 1.00 2010/05/26 H
Ver, 1.01 2018/10/07 WinInfo BGetWindowFileName, FindWindowFromProcess, GetFrameRect ǉB
Ver, 1.02 2018/10/21 FindWindowFromProcess  FindWindowIn ɉB
Ver, 1.03 2018/11/08 GetFrameRect CBGetParentOnly, IsTopLevelWindow, MoveFrame ǉB
Ver, 1.04 2018/11/18 MatchesWindow ǉBǂǂ񑝂ccB

EBhE֘A[`WB
ɂăo[Wԍɂ͈ӖȂB

*)

unit WndInfo;

interface

uses
  Winapi.Windows, System.SysUtils, System.Masks, Winapi.PsAPI, Winapi.Dwmapi;

function GetWindowTextStr(Wnd: HWND): string;
function GetClassNameStr(Wnd: HWND): string;
function MatchesWindow(Wnd: HWND; ClassName, WindowName: string): Boolean;
function GetWindowFileName(Wnd: HWND): string;
function FindWindowIn(Process: Cardinal; ClassName, WindowName: string): HWND;
function GetParentOnly(Wnd: HWND): HWND;
function IsTopLevelWindow(Wnd: HWND): Boolean;
function GetFrameRect(Wnd: HWND; var ARect: TRect): Boolean;
function MoveFrame(Wnd: HWND; X, Y, Width, Height: Integer; Repaint: Boolean): Boolean;

implementation

{ EBhE^Cg𓾂 }
function GetWindowTextStr(Wnd: HWND): string;
var
  Len: Integer;
  Buf: PChar;
begin
  Result := '';
  Buf := StrAlloc(MAX_PATH);
  try
    Len := GetWindowText(Wnd, Buf, StrBufSize(Buf));
    SetString(Result, Buf, Len);
  finally
    StrDispose(Buf);
  end;
end;

{ NX𓾂 }
function GetClassNameStr(Wnd: HWND): string;
var
  Len: Integer;
  Buf: PChar;
begin
  Result := '';
  Buf := StrAlloc(MAX_PATH);
  try
    Len := GetClassName(Wnd, Buf, StrBufSize(Buf));
    SetString(Result, Buf, Len);
  finally
    StrDispose(Buf);
  end;
end;

{ EBhEw̃NXƃ^CgɈv邩 }
function MatchesWindow(Wnd: HWND; ClassName, WindowName: string): Boolean;
var
  Name,
  Title: string;
begin
  Name := GetClassNameStr(Wnd);
  Title := GetWindowTextStr(Wnd);
  Result := MatchesMask(Name, ClassName) and MatchesMask(Title, WindowName);
end;

{ EBhEnhst@C𓾂 }
function GetWindowFileName(Wnd: HWND): string;
var
  ID: DWORD;
  Handle: HWND;
  Path: array[0..MAX_PATH] of Char;
begin
  GetWindowThreadProcessId(Wnd, @ID);
  Handle := OpenProcess(PROCESS_QUERY_INFORMATION or PROCESS_VM_READ, False, ID);
  try
    if GetModuleFileNameEx(Handle, 0, Path, SizeOf(Path)) > 0 then
      Result := Path
    else
      Result := ''; //! SysErrorMessage(GetLastError);
  finally
    CloseHandle(Handle);
  end;
end;

var
  //FindWindowIn p
  ProcessID: Cardinal;
  ProcessName,
  ProcessTitle: string;

//FindWindowIn R[obNB
function EnumWindowsProc(Wnd: HWND; Param: LPARAM): BOOL; stdcall;
var
  ID: Cardinal;
begin
  GetWindowThreadProcessId(Wnd, @ID);
  Result := not ((ID = ProcessID) and MatchesWindow(Wnd, ProcessName, ProcessTitle));
  if not Result then
    LPDWORD(Param)^ := Wnd;
end;

{ vZXIDɑEBhEnh }
function FindWindowIn(Process: Cardinal; ClassName, WindowName: string): HWND;
begin
  ProcessID := Process;
  ProcessName := ClassName;
  ProcessTitle := WindowName;

  Result := 0;
  EnumWindows(@EnumWindowsProc, LPARAM(@Result));
end;

{ eEBhẼnĥ݁iI[i[EBhE͏jԂ }
function GetParentOnly(Wnd: HWND): HWND;
begin
  Result := GetAncestor(Wnd, GA_PARENT);
  if Result = GetDesktopWindow then
    Result := 0;
end;

{ gbvxEBhEׂ
  J API 炵ǁAWin7 ȑȎΉsȂ̂ŎB
  The Undocumented IsTopLevelWindow API
  http://jellevergeer.com/the-undocumented-istoplevelwindow-api/ }
function IsTopLevelWindow(Wnd: HWND): Boolean;
begin
  Result := GetDesktopWindow = GetAncestor(Wnd, GA_PARENT);
end;

{ EBhE̊O𓾂 }
function GetFrameRect(Wnd: HWND; var ARect: TRect): Boolean;
begin
  //gbvxEBhEȂ E_HANDLE Ԃ݂B
  Result := IsTopLevelWindow(Wnd) and TOSVersion.Check(6) and DwmCompositionEnabled;
  if Result then
    Result := DwmGetWindowAttribute(Wnd, DWMWA_EXTENDED_FRAME_BOUNDS, @ARect, SizeOf(TRect)) = S_OK;
  if not Result then
    Result := GetWindowRect(Wnd, ARect);
end;

{ EBhĚ̈ʒuƃTCYύX }
function MoveFrame(Wnd: HWND; X, Y, Width, Height: Integer; Repaint: Boolean): Boolean;
var
  Frame, Outer, Inner: TRect;
begin
  Frame := TRect.Create(X, Y, X + Width, Y + Height);
  GetWindowRect(Wnd, Outer);
  GetFrameRect(Wnd, Inner);
  Frame.Inflate(
    Inner.Left - Outer.Left,
    Inner.Top - Outer.Top,
    Outer.Right - Inner.Right,
    Outer.Bottom - Inner.Bottom
  );
  Result := MoveWindow(Wnd, Frame.Left, Frame.Top, Frame.Width, Frame.Height, Repaint);
end;

end.
